from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator


##


_GLOBAL_BREAKPOINT = ContextVar("GLOBAL_BREAKPOINT", default=False)


def global_breakpoint() -> None:
    """Set a breakpoint if the global breakpoint is enabled."""
    if _GLOBAL_BREAKPOINT.get():  # pragma: no cover
        breakpoint()  # noqa: T100


def set_global_breakpoint() -> None:
    """Set the global breakpoint ."""
    _ = _GLOBAL_BREAKPOINT.set(True)


##


@contextmanager
def yield_set_context(var: ContextVar[bool], /) -> Iterator[None]:
    """Yield a context var as being set."""
    token = var.set(True)
    try:
        yield
    finally:
        _ = var.reset(token)


__all__ = ["global_breakpoint", "set_global_breakpoint", "yield_set_context"]



from __future__ import annotations

from zoneinfo import ZoneInfo

from hypothesis import given
from hypothesis.strategies import sampled_from

from utilities.tzdata import HongKong, Tokyo, USCentral, USEastern


class TestTimeZones:
    @given(time_zone=sampled_from([HongKong, Tokyo, USCentral, USEastern]))
    def test_main(self, *, time_zone: ZoneInfo) -> None:
        assert isinstance(time_zone, ZoneInfo)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from hypothesis import given
from hypothesis.strategies import DataObject, data, sampled_from
from pytest import mark, param, raises

from utilities.sentinel import (
    SENTINEL_REPR,
    ParseSentinelError,
    Sentinel,
    is_sentinel,
    parse_sentinel,
    sentinel,
)

if TYPE_CHECKING:
    from collections.abc import Callable


class TestIsSentinel:
    @mark.parametrize(("obj", "expected"), [param(None, False), param(sentinel, True)])
    def test_main(self, *, obj: Any, expected: bool) -> None:
        assert is_sentinel(obj) is expected


class TestParseSentinel:
    @given(data=data())
    def test_main(self, *, data: DataObject) -> None:
        text = str(sentinel)
        text_use = data.draw(sampled_from(["", text, text.lower(), text.upper()]))
        result = parse_sentinel(text_use)
        assert result is sentinel

    @given(text=sampled_from(["invalid", "ssentinell"]))
    def test_error(self, *, text: str) -> None:
        with raises(
            ParseSentinelError, match="Unable to parse sentinel value; got '.*'"
        ):
            _ = parse_sentinel(text)


class TestSentinel:
    def test_isinstance(self) -> None:
        assert isinstance(sentinel, Sentinel)

    @mark.parametrize("method", [param(repr), param(str)])
    def test_repr_and_str(self, method: Callable[..., str]) -> None:
        assert method(sentinel) == SENTINEL_REPR

    def test_singletone(self) -> None:
        assert Sentinel() is sentinel



from __future__ import annotations

from re import search
from time import sleep
from typing import TYPE_CHECKING

from utilities.pyinstrument import profile

if TYPE_CHECKING:
    from pathlib import Path


class TestProfile:
    def test_main(self, tmp_path: Path) -> None:
        with profile(tmp_path):
            sleep(1e-3)

        (file,) = tmp_path.iterdir()
        assert search(r"^profile__[\dT]+\.html$", file.name)



from __future__ import annotations

from collections import defaultdict
from collections.abc import (
    AsyncIterator,
    Callable,
    Hashable,
    Iterable,
    Iterator,
    Sequence,
    Sized,
)
from collections.abc import Set as AbstractSet
from contextlib import asynccontextmanager
from dataclasses import dataclass
from functools import reduce
from itertools import chain
from math import floor
from operator import ge, le
from re import search
from socket import gaierror
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypeGuard,
    assert_never,
    cast,
    overload,
    override,
)

import sqlalchemy
from sqlalchemy import (
    URL,
    Column,
    Connection,
    Engine,
    Insert,
    Selectable,
    Table,
    and_,
    case,
    insert,
    select,
    text,
)
from sqlalchemy.dialects.mssql import dialect as mssql_dialect
from sqlalchemy.dialects.mysql import dialect as mysql_dialect
from sqlalchemy.dialects.oracle import dialect as oracle_dialect
from sqlalchemy.dialects.postgresql import dialect as postgresql_dialect
from sqlalchemy.dialects.postgresql import insert as postgresql_insert
from sqlalchemy.dialects.postgresql.asyncpg import PGDialect_asyncpg
from sqlalchemy.dialects.postgresql.psycopg import PGDialect_psycopg
from sqlalchemy.dialects.sqlite import dialect as sqlite_dialect
from sqlalchemy.dialects.sqlite import insert as sqlite_insert
from sqlalchemy.exc import ArgumentError, DatabaseError
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, create_async_engine
from sqlalchemy.orm import (
    DeclarativeBase,
    InstrumentedAttribute,
    class_mapper,
    declared_attr,
)
from sqlalchemy.orm.exc import UnmappedClassError
from sqlalchemy.pool import NullPool, Pool

from utilities.asyncio import timeout_td
from utilities.functions import (
    ensure_str,
    get_class_name,
    is_sequence_of_tuple_or_str_mapping,
    is_string_mapping,
    is_tuple,
    is_tuple_or_str_mapping,
    yield_object_attributes,
)
from utilities.iterables import (
    CheckLengthError,
    CheckSubSetError,
    OneEmptyError,
    OneNonUniqueError,
    check_length,
    check_subset,
    chunked,
    merge_sets,
    merge_str_mappings,
    one,
)
from utilities.os import is_pytest
from utilities.reprlib import get_repr
from utilities.text import secret_str, snake_case
from utilities.types import (
    Delta,
    MaybeIterable,
    MaybeType,
    StrMapping,
    TupleOrStrMapping,
)

if TYPE_CHECKING:
    from enum import Enum, StrEnum


type EngineOrConnectionOrAsync = Engine | Connection | AsyncEngine | AsyncConnection
type Dialect = Literal["mssql", "mysql", "oracle", "postgresql", "sqlite"]
type DialectOrEngineOrConnectionOrAsync = Dialect | EngineOrConnectionOrAsync
type ORMInstOrClass = DeclarativeBase | type[DeclarativeBase]
type TableOrORMInstOrClass = Table | ORMInstOrClass
CHUNK_SIZE_FRAC = 0.8


##


_SELECT = text("SELECT 1")


def check_connect(engine: Engine, /) -> bool:
    """Check if an engine can connect."""
    try:
        with engine.connect() as conn:
            return bool(conn.execute(_SELECT).scalar_one())
    except (gaierror, ConnectionRefusedError, DatabaseError):  # pragma: no cover
        return False


async def check_connect_async(
    engine: AsyncEngine,
    /,
    *,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
) -> bool:
    """Check if an engine can connect."""
    try:
        async with timeout_td(timeout, error=error), engine.connect() as conn:
            return bool((await conn.execute(_SELECT)).scalar_one())
    except (gaierror, ConnectionRefusedError, DatabaseError, TimeoutError):
        return False


##


async def check_engine(
    engine: AsyncEngine,
    /,
    *,
    timeout: Delta | None = None,
    error: MaybeType[Exception] = TimeoutError,
    num_tables: int | tuple[int, float] | None = None,
) -> None:
    """Check that an engine can connect.

    Optionally query for the number of tables, or the number of columns in
    such a table.
    """
    match _get_dialect(engine):
        case "mssql" | "mysql" | "postgresql":  # skipif-ci-and-not-linux
            query = "select * from information_schema.tables"
        case "oracle":  # pragma: no cover
            query = "select * from all_objects"
        case "sqlite":
            query = "select * from sqlite_master where type='table'"
        case never:
            assert_never(never)
    statement = text(query)
    async with yield_connection(engine, timeout=timeout, error=error) as conn:
        rows = (await conn.execute(statement)).all()
    if num_tables is not None:
        try:
            check_length(rows, equal_or_approx=num_tables)
        except CheckLengthError as err:
            raise CheckEngineError(
                engine=engine, rows=err.obj, expected=num_tables
            ) from None


@dataclass(kw_only=True, slots=True)
class CheckEngineError(Exception):
    engine: AsyncEngine
    rows: Sized
    expected: int | tuple[int, float]

    @override
    def __str__(self) -> str:
        return f"{get_repr(self.engine)} must have {self.expected} table(s); got {len(self.rows)}"


##


def columnwise_max(*columns: Any) -> Any:
    """Compute the columnwise max of a number of columns."""
    return _columnwise_minmax(*columns, op=ge)


def columnwise_min(*columns: Any) -> Any:
    """Compute the columnwise min of a number of columns."""
    return _columnwise_minmax(*columns, op=le)


def _columnwise_minmax(*columns: Any, op: Callable[[Any, Any], Any]) -> Any:
    """Compute the columnwise min of a number of columns."""

    def func(x: Any, y: Any, /) -> Any:
        x_none = x.is_(None)
        y_none = y.is_(None)
        col = case(
            (and_(x_none, y_none), None),
            (and_(~x_none, y_none), x),
            (and_(x_none, ~y_none), y),
            (op(x, y), x),
            else_=y,
        )
        # try auto-label
        names = {
            value for col in [x, y] if (value := getattr(col, "name", None)) is not None
        }
        try:
            (name,) = names
        except ValueError:
            return col
        else:
            return col.label(name)

    return reduce(func, columns)


##


@overload
def create_engine(
    drivername: str,
    /,
    *,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    database: str | None = None,
    query: StrMapping | None = None,
    poolclass: type[Pool] | None = NullPool,
    async_: Literal[True],
) -> AsyncEngine: ...
@overload
def create_engine(
    drivername: str,
    /,
    *,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    database: str | None = None,
    query: StrMapping | None = None,
    poolclass: type[Pool] | None = NullPool,
    async_: Literal[False] = False,
) -> Engine: ...
@overload
def create_engine(
    drivername: str,
    /,
    *,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    database: str | None = None,
    query: StrMapping | None = None,
    poolclass: type[Pool] | None = NullPool,
    async_: bool = False,
) -> Engine | AsyncEngine: ...
def create_engine(
    drivername: str,
    /,
    *,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    database: str | None = None,
    query: StrMapping | None = None,
    poolclass: type[Pool] | None = NullPool,
    async_: bool = False,
) -> Engine | AsyncEngine:
    """Create a SQLAlchemy engine."""
    if query is None:
        kwargs = {}
    else:

        def func(x: MaybeIterable[str], /) -> list[str] | str:
            return x if isinstance(x, str) else list(x)

        kwargs = {"query": {k: func(v) for k, v in query.items()}}
    url = URL.create(
        drivername,
        username=username,
        password=password,
        host=host,
        port=port,
        database=database,
        **kwargs,
    )
    match async_:
        case False:
            return sqlalchemy.create_engine(url, poolclass=poolclass)
        case True:
            return create_async_engine(url, poolclass=poolclass)
        case never:
            assert_never(never)


##


async def ensure_database_created(super_: URL, database: str, /) -> None:
    """Ensure a database is created."""
    engine = create_async_engine(super_, isolation_level="AUTOCOMMIT")
    async with engine.begin() as conn:
        try:
            _ = await conn.execute(text(f"CREATE DATABASE {database}"))
        except DatabaseError as error:
            _ensure_tables_maybe_reraise(error, 'database ".*" already exists')


async def ensure_database_dropped(super_: URL, database: str, /) -> None:
    """Ensure a database is dropped."""
    engine = create_async_engine(super_, isolation_level="AUTOCOMMIT")
    async with engine.begin() as conn:
        _ = await conn.execute(text(f"DROP DATABASE IF EXISTS {database}"))


##


async def ensure_tables_created(
    engine: AsyncEngine,
    /,
    *tables_or_orms: TableOrORMInstOrClass,
    timeout: Delta | None = None,
    error: MaybeType[Exception] = TimeoutError,
) -> None:
    """Ensure a table/set of tables is/are created."""
    tables = set(map(get_table, tables_or_orms))
    match dialect := _get_dialect(engine):
        case "mysql":  # pragma: no cover
            raise NotImplementedError(dialect)
        case "postgresql":  # skipif-ci-and-not-linux
            match = "relation .* already exists"
        case "mssql":  # pragma: no cover
            match = "There is already an object named .* in the database"
        case "oracle":  # pragma: no cover
            match = "ORA-00955: name is already used by an existing object"
        case "sqlite":
            match = "table .* already exists"
        case never:
            assert_never(never)
    async with yield_connection(engine, timeout=timeout, error=error) as conn:
        for table in tables:
            try:
                await conn.run_sync(table.create)
            except DatabaseError as err:
                _ensure_tables_maybe_reraise(err, match)


async def ensure_tables_dropped(
    engine: AsyncEngine,
    *tables_or_orms: TableOrORMInstOrClass,
    timeout: Delta | None = None,
    error: MaybeType[Exception] = TimeoutError,
) -> None:
    """Ensure a table/set of tables is/are dropped."""
    tables = set(map(get_table, tables_or_orms))
    match dialect := _get_dialect(engine):
        case "mysql":  # pragma: no cover
            raise NotImplementedError(dialect)
        case "postgresql":  # skipif-ci-and-not-linux
            match = "table .* does not exist"
        case "mssql":  # pragma: no cover
            match = "Cannot drop the table .*, because it does not exist or you do not have permission"
        case "oracle":  # pragma: no cover
            match = "ORA-00942: table or view does not exist"
        case "sqlite":
            match = "no such table"
        case never:
            assert_never(never)
    async with yield_connection(engine, timeout=timeout, error=error) as conn:
        for table in tables:
            try:
                await conn.run_sync(table.drop)
            except DatabaseError as err:
                _ensure_tables_maybe_reraise(err, match)


##


def enum_name(enum: type[Enum], /) -> str:
    """Get the name of an Enum."""
    return f"{snake_case(get_class_name(enum))}_enum"


##


def enum_values(enum: type[StrEnum], /) -> list[str]:
    """Get the values of a StrEnum."""
    return [e.value for e in enum]


##


@dataclass(kw_only=True, slots=True)
class ExtractURLOutput:
    username: str
    password: secret_str
    host: str
    port: int
    database: str


def extract_url(url: URL, /) -> ExtractURLOutput:
    """Extract the database, host & port from a URL."""
    if url.username is None:
        raise _ExtractURLUsernameError(url=url)
    if url.password is None:
        raise _ExtractURLPasswordError(url=url)
    if url.host is None:
        raise _ExtractURLHostError(url=url)
    if url.port is None:
        raise _ExtractURLPortError(url=url)
    if url.database is None:
        raise _ExtractURLDatabaseError(url=url)
    return ExtractURLOutput(
        username=url.username,
        password=secret_str(url.password),
        host=url.host,
        port=url.port,
        database=url.database,
    )


@dataclass(kw_only=True, slots=True)
class ExtractURLError(Exception):
    url: URL


@dataclass(kw_only=True, slots=True)
class _ExtractURLUsernameError(ExtractURLError):
    @override
    def __str__(self) -> str:
        return f"Expected URL to contain a user name; got {self.url}"


@dataclass(kw_only=True, slots=True)
class _ExtractURLPasswordError(ExtractURLError):
    @override
    def __str__(self) -> str:
        return f"Expected URL to contain a password; got {self.url}"


@dataclass(kw_only=True, slots=True)
class _ExtractURLHostError(ExtractURLError):
    @override
    def __str__(self) -> str:
        return f"Expected URL to contain a host; got {self.url}"


@dataclass(kw_only=True, slots=True)
class _ExtractURLPortError(ExtractURLError):
    @override
    def __str__(self) -> str:
        return f"Expected URL to contain a port; got {self.url}"


@dataclass(kw_only=True, slots=True)
class _ExtractURLDatabaseError(ExtractURLError):
    @override
    def __str__(self) -> str:
        return f"Expected URL to contain a database; got {self.url}"


##


def get_chunk_size(
    dialect_or_engine_or_conn: DialectOrEngineOrConnectionOrAsync,
    table_or_orm_or_num_cols: TableOrORMInstOrClass | Sized | int,
    /,
    *,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
) -> int:
    """Get the maximum chunk size for an engine."""
    max_params = _get_dialect_max_params(dialect_or_engine_or_conn)
    match table_or_orm_or_num_cols:
        case Table() | DeclarativeBase() | type() as table_or_orm:
            return get_chunk_size(
                dialect_or_engine_or_conn,
                get_columns(table_or_orm),
                chunk_size_frac=chunk_size_frac,
            )
        case Sized() as sized:
            return get_chunk_size(
                dialect_or_engine_or_conn, len(sized), chunk_size_frac=chunk_size_frac
            )
        case int() as num_cols:
            size = floor(chunk_size_frac * max_params / num_cols)
            return max(size, 1)
        case never:
            assert_never(never)


##


def get_column_names(table_or_orm: TableOrORMInstOrClass, /) -> list[str]:
    """Get the column names from a table or ORM instance/class."""
    return [col.name for col in get_columns(table_or_orm)]


##


def get_columns(table_or_orm: TableOrORMInstOrClass, /) -> list[Column[Any]]:
    """Get the columns from a table or ORM instance/class."""
    return list(get_table(table_or_orm).columns)


##


def get_primary_key_values(orm: DeclarativeBase, /) -> tuple[Any, ...]:
    """Get a tuple of the primary key values."""
    return tuple(getattr(orm, c.name) for c in yield_primary_key_columns(orm))


##


def get_table(table_or_orm: TableOrORMInstOrClass, /) -> Table:
    """Get the table from a Table or mapped class."""
    if isinstance(table_or_orm, Table):
        return table_or_orm
    if is_orm(table_or_orm):
        return cast("Table", table_or_orm.__table__)
    raise GetTableError(obj=table_or_orm)


@dataclass(kw_only=True, slots=True)
class GetTableError(Exception):
    obj: Any

    @override
    def __str__(self) -> str:
        return f"Object {self.obj} must be a Table or mapped class; got {get_class_name(self.obj)!r}"


##


def get_table_name(table_or_orm: TableOrORMInstOrClass, /) -> str:
    """Get the table name from a Table or mapped class."""
    return get_table(table_or_orm).name


##


def hash_primary_key_values(orm: DeclarativeBase, /) -> int:
    """Compute a hash of the primary key values."""
    return hash(get_primary_key_values(orm))


##


type _PairOfTupleAndTable = tuple[tuple[Any, ...], TableOrORMInstOrClass]
type _PairOfStrMappingAndTable = tuple[StrMapping, TableOrORMInstOrClass]
type _PairOfTupleOrStrMappingAndTable = tuple[TupleOrStrMapping, TableOrORMInstOrClass]
type _PairOfSequenceOfTupleOrStrMappingAndTable = tuple[
    Sequence[TupleOrStrMapping], TableOrORMInstOrClass
]
type _InsertItem = (
    _PairOfTupleOrStrMappingAndTable
    | _PairOfSequenceOfTupleOrStrMappingAndTable
    | DeclarativeBase
    | Sequence[_PairOfTupleOrStrMappingAndTable]
    | Sequence[DeclarativeBase]
)
type _NormalizedItem = tuple[Table, StrMapping]
type _InsertPair = tuple[Table, Sequence[StrMapping]]
type _InsertTriple = tuple[Table, Insert, Sequence[StrMapping] | None]


async def insert_items(
    engine: AsyncEngine,
    *items: _InsertItem,
    snake: bool = False,
    is_upsert: bool = False,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
    timeout_create: Delta | None = None,
    error_create: MaybeType[Exception] = TimeoutError,
    timeout_insert: Delta | None = None,
    error_insert: MaybeType[Exception] = TimeoutError,
) -> None:
    """Insert a set of items into a database.

    These can be one of the following:
     - pair of tuple & table/class:           (x1, x2, ...), table_cls
     - pair of dict & table/class:            {k1=v1, k2=v2, ...), table_cls
     - pair of list of tuples & table/class:  [(x11, x12, ...),
                                               (x21, x22, ...),
                                               ...], table_cls
     - pair of list of dicts & table/class:   [{k1=v11, k2=v12, ...},
                                               {k1=v21, k2=v22, ...},
                                               ...], table/class
     - list of pairs of tuple & table/class:  [((x11, x12, ...), table_cls1),
                                               ((x21, x22, ...), table_cls2),
                                               ...]
     - list of pairs of dict & table/class:   [({k1=v11, k2=v12, ...}, table_cls1),
                                               ({k1=v21, k2=v22, ...}, table_cls2),
                                               ...]
     - mapped class:                          Obj(k1=v1, k2=v2, ...)
     - list of mapped classes:                [Obj(k1=v11, k2=v12, ...),
                                               Obj(k1=v21, k2=v22, ...),
                                               ...]
    """
    normalized = chain.from_iterable(
        _insert_items_yield_normalized(i, snake=snake) for i in items
    )
    triples = _insert_items_yield_triples(
        engine, normalized, is_upsert=is_upsert, chunk_size_frac=chunk_size_frac
    )
    if not assume_tables_exist:
        triples = list(triples)
        tables = {table for table, _, _ in triples}
        await ensure_tables_created(
            engine, *tables, timeout=timeout_create, error=error_create
        )
    for _, ins, parameters in triples:
        async with yield_connection(
            engine, timeout=timeout_insert, error=error_insert
        ) as conn:
            _ = await conn.execute(ins, parameters=parameters)


def _insert_items_yield_normalized(
    item: _InsertItem, /, *, snake: bool = False
) -> Iterator[_NormalizedItem]:
    if _is_pair_of_str_mapping_and_table(item):
        mapping, table_or_orm = item
        adjusted = _map_mapping_to_table(mapping, table_or_orm, snake=snake)
        yield (get_table(table_or_orm), adjusted)
        return
    if _is_pair_of_tuple_and_table(item):
        tuple_, table_or_orm = item
        mapping = _tuple_to_mapping(tuple_, table_or_orm)
        yield from _insert_items_yield_normalized((mapping, table_or_orm), snake=snake)
        return
    if _is_pair_of_sequence_of_tuple_or_string_mapping_and_table(item):
        items, table_or_orm = item
        pairs = [(i, table_or_orm) for i in items]
        for p in pairs:
            yield from _insert_items_yield_normalized(p, snake=snake)
        return
    if isinstance(item, DeclarativeBase):
        mapping = _orm_inst_to_dict(item)
        yield from _insert_items_yield_normalized((mapping, item), snake=snake)
        return
    try:
        _ = iter(item)
    except TypeError:
        raise InsertItemsError(item=item) from None
    if all(map(_is_pair_of_tuple_or_str_mapping_and_table, item)):
        pairs = cast("Sequence[_PairOfTupleOrStrMappingAndTable]", item)
        for p in pairs:
            yield from _insert_items_yield_normalized(p, snake=snake)
        return
    if all(map(is_orm, item)):
        classes = cast("Sequence[DeclarativeBase]", item)
        for c in classes:
            yield from _insert_items_yield_normalized(c, snake=snake)
        return
    raise InsertItemsError(item=item)


def _insert_items_yield_triples(
    engine: AsyncEngine,
    items: Iterable[_NormalizedItem],
    /,
    *,
    is_upsert: bool = False,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
) -> Iterable[_InsertTriple]:
    pairs = _insert_items_yield_chunked_pairs(
        engine, items, is_upsert=is_upsert, chunk_size_frac=chunk_size_frac
    )
    for table, mappings in pairs:
        match is_upsert, _get_dialect(engine):
            case False, "oracle":  # pragma: no cover
                ins = insert(table)
                parameters = mappings
            case False, _:
                ins = insert(table).values(mappings)
                parameters = None
            case True, _:
                ins = _insert_items_build_insert_with_on_conflict_do_update(
                    engine, table, mappings
                )
                parameters = None
            case never:
                assert_never(never)
        yield table, ins, parameters


def _insert_items_yield_chunked_pairs(
    engine: AsyncEngine,
    items: Iterable[_NormalizedItem],
    /,
    *,
    is_upsert: bool = False,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
) -> Iterable[_InsertPair]:
    for table, mappings in _insert_items_yield_raw_pairs(items, is_upsert=is_upsert):
        chunk_size = get_chunk_size(engine, table, chunk_size_frac=chunk_size_frac)
        for mappings_i in chunked(mappings, chunk_size):
            yield table, list(mappings_i)


def _insert_items_yield_raw_pairs(
    items: Iterable[_NormalizedItem], /, *, is_upsert: bool = False
) -> Iterable[_InsertPair]:
    by_table: defaultdict[Table, list[StrMapping]] = defaultdict(list)
    for table, mapping in items:
        by_table[table].append(mapping)
    for table, mappings in by_table.items():
        yield from _insert_items_yield_raw_pairs_one(
            table, mappings, is_upsert=is_upsert
        )


def _insert_items_yield_raw_pairs_one(
    table: Table, mappings: Iterable[StrMapping], /, *, is_upsert: bool = False
) -> Iterable[_InsertPair]:
    merged = _insert_items_yield_merged_mappings(table, mappings)
    match is_upsert:
        case True:
            by_keys: defaultdict[frozenset[str], list[StrMapping]] = defaultdict(list)
            for mapping in merged:
                non_null = {k: v for k, v in mapping.items() if v is not None}
                by_keys[frozenset(non_null)].append(non_null)
            for mappings_i in by_keys.values():
                yield table, mappings_i
        case False:
            yield table, list(merged)
        case never:
            assert_never(never)


def _insert_items_yield_merged_mappings(
    table: Table, mappings: Iterable[StrMapping], /
) -> Iterable[StrMapping]:
    columns = list(yield_primary_key_columns(table))
    col_names = [c.name for c in columns]
    cols_auto = {c.name for c in columns if c.autoincrement in {True, "auto"}}
    cols_non_auto = set(col_names) - cols_auto
    by_key: defaultdict[tuple[Hashable, ...], list[StrMapping]] = defaultdict(list)
    for mapping in mappings:
        check_subset(cols_non_auto, mapping)
        has_all_auto = set(cols_auto).issubset(mapping)
        if has_all_auto:
            pkey = tuple(mapping[k] for k in col_names)
            rest: StrMapping = {k: v for k, v in mapping.items() if k not in col_names}
            by_key[pkey].append(rest)
        else:
            yield mapping
    for k, v in by_key.items():
        head = dict(zip(col_names, k, strict=True))
        yield merge_str_mappings(head, *v)


def _insert_items_build_insert_with_on_conflict_do_update(
    engine: AsyncEngine, table: Table, mappings: Iterable[StrMapping], /
) -> Insert:
    primary_key = cast("Any", table.primary_key)
    mappings = list(mappings)
    columns = merge_sets(*mappings)
    match _get_dialect(engine):
        case "postgresql":  # skipif-ci-and-not-linux
            ins = postgresql_insert(table).values(mappings)
            set_ = {c: getattr(ins.excluded, c) for c in columns}
            return ins.on_conflict_do_update(constraint=primary_key, set_=set_)
        case "sqlite":
            ins = sqlite_insert(table).values(mappings)
            set_ = {c: getattr(ins.excluded, c) for c in columns}
            return ins.on_conflict_do_update(index_elements=primary_key, set_=set_)
        case "mssql" | "mysql" | "oracle" as dialect:  # pragma: no cover
            raise NotImplementedError(dialect)
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class InsertItemsError(Exception):
    item: _InsertItem

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


##


def is_orm(obj: Any, /) -> TypeGuard[ORMInstOrClass]:
    """Check if an object is an ORM instance/class."""
    if isinstance(obj, type):
        try:
            _ = class_mapper(cast("Any", obj))
        except (ArgumentError, UnmappedClassError):
            return False
        return True
    return is_orm(type(obj))


##


def is_table_or_orm(obj: Any, /) -> TypeGuard[TableOrORMInstOrClass]:
    """Check if an object is a Table or an ORM instance/class."""
    return isinstance(obj, Table) or is_orm(obj)


##


async def migrate_data(
    table_or_orm_from: TableOrORMInstOrClass,
    engine_from: AsyncEngine,
    engine_to: AsyncEngine,
    /,
    *,
    table_or_orm_to: TableOrORMInstOrClass | None = None,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
    timeout_create: Delta | None = None,
    error_create: MaybeType[Exception] = TimeoutError,
    timeout_insert: Delta | None = None,
    error_insert: MaybeType[Exception] = TimeoutError,
) -> None:
    """Migrate the contents of a table from one database to another."""
    table_from = get_table(table_or_orm_from)
    async with engine_from.begin() as conn:
        rows = (await conn.execute(select(table_from))).all()
    mappings = [dict(r._mapping) for r in rows]  # noqa: SLF001
    table_to = table_from if table_or_orm_to is None else get_table(table_or_orm_to)
    items = (mappings, table_to)
    await insert_items(
        engine_to,
        items,
        chunk_size_frac=chunk_size_frac,
        assume_tables_exist=assume_tables_exist,
        timeout_create=timeout_create,
        error_create=error_create,
        timeout_insert=timeout_insert,
        error_insert=error_insert,
    )


##


def selectable_to_string(
    selectable: Selectable[Any], engine_or_conn: EngineOrConnectionOrAsync, /
) -> str:
    """Convert a selectable into a string."""
    com = selectable.compile(
        dialect=engine_or_conn.dialect, compile_kwargs={"literal_binds": True}
    )
    return str(com)


##


class TablenameMixin:
    """Mix-in for an auto-generated tablename."""

    @cast("Any", declared_attr)
    def __tablename__(cls) -> str:  # noqa: N805
        return snake_case(get_class_name(cls))


##


@asynccontextmanager
async def yield_connection(
    engine: AsyncEngine,
    /,
    *,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
) -> AsyncIterator[AsyncConnection]:
    """Yield an async connection."""
    try:
        async with timeout_td(timeout, error=error), engine.begin() as conn:
            yield conn
    except GeneratorExit:  # pragma: no cover
        if not is_pytest():
            raise
        return


##


def yield_primary_key_columns(
    obj: TableOrORMInstOrClass,
    /,
    *,
    autoincrement: bool | Literal["auto", "ignore_fk"] | None = None,
) -> Iterator[Column]:
    """Yield the primary key columns of a table."""
    table = get_table(obj)
    for column in table.primary_key:
        if (autoincrement is None) or (autoincrement == column.autoincrement):
            yield column


##


def _ensure_tables_maybe_reraise(error: DatabaseError, match: str, /) -> None:
    """Re-raise the error if it does not match the required statement."""
    if not search(match, ensure_str(one(error.args))):
        raise error  # pragma: no cover


##


def _get_dialect(engine_or_conn: EngineOrConnectionOrAsync, /) -> Dialect:
    """Get the dialect of a database."""
    dialect = engine_or_conn.dialect
    if isinstance(dialect, mssql_dialect):  # pragma: no cover
        return "mssql"
    if isinstance(dialect, mysql_dialect):  # pragma: no cover
        return "mysql"
    if isinstance(dialect, oracle_dialect):  # pragma: no cover
        return "oracle"
    if isinstance(  # skipif-ci-and-not-linux
        dialect, (postgresql_dialect, PGDialect_asyncpg, PGDialect_psycopg)
    ):
        return "postgresql"
    if isinstance(dialect, sqlite_dialect):
        return "sqlite"
    msg = f"Unknown dialect: {dialect}"  # pragma: no cover
    raise NotImplementedError(msg)  # pragma: no cover


##


def _get_dialect_max_params(
    dialect_or_engine_or_conn: DialectOrEngineOrConnectionOrAsync, /
) -> int:
    """Get the max number of parameters of a dialect."""
    match dialect_or_engine_or_conn:
        case "mssql":  # pragma: no cover
            return 2100
        case "mysql":  # pragma: no cover
            return 65535
        case "oracle":  # pragma: no cover
            return 1000
        case "postgresql":  # skipif-ci-and-not-linux
            return 32767
        case "sqlite":
            return 100
        case (
            Engine()
            | Connection()
            | AsyncEngine()
            | AsyncConnection() as engine_or_conn
        ):
            dialect = _get_dialect(engine_or_conn)
            return _get_dialect_max_params(dialect)
        case never:
            assert_never(never)


##


def _is_pair_of_sequence_of_tuple_or_string_mapping_and_table(
    obj: Any, /
) -> TypeGuard[_PairOfSequenceOfTupleOrStrMappingAndTable]:
    """Check if an object is a pair of a sequence of tuples/string mappings and a table."""
    return _is_pair_with_predicate_and_table(obj, is_sequence_of_tuple_or_str_mapping)


def _is_pair_of_str_mapping_and_table(
    obj: Any, /
) -> TypeGuard[_PairOfStrMappingAndTable]:
    """Check if an object is a pair of a string mapping and a table."""
    return _is_pair_with_predicate_and_table(obj, is_string_mapping)


def _is_pair_of_tuple_and_table(obj: Any, /) -> TypeGuard[_PairOfTupleAndTable]:
    """Check if an object is a pair of a tuple and a table."""
    return _is_pair_with_predicate_and_table(obj, is_tuple)


def _is_pair_of_tuple_or_str_mapping_and_table(
    obj: Any, /
) -> TypeGuard[_PairOfTupleOrStrMappingAndTable]:
    """Check if an object is a pair of a tuple/string mapping and a table."""
    return _is_pair_with_predicate_and_table(obj, is_tuple_or_str_mapping)


def _is_pair_with_predicate_and_table[T](
    obj: Any, predicate: Callable[[Any], TypeGuard[T]], /
) -> TypeGuard[tuple[T, TableOrORMInstOrClass]]:
    """Check if an object is pair and a table."""
    return (
        isinstance(obj, tuple)
        and (len(obj) == 2)
        and predicate(obj[0])
        and is_table_or_orm(obj[1])
    )


##


def _map_mapping_to_table(
    mapping: StrMapping, table_or_orm: TableOrORMInstOrClass, /, *, snake: bool = False
) -> StrMapping:
    """Map a mapping to a table."""
    columns = get_column_names(table_or_orm)
    if not snake:
        try:
            check_subset(mapping, columns)
        except CheckSubSetError as error:
            raise _MapMappingToTableExtraColumnsError(
                mapping=mapping, columns=columns, extra=error.extra
            ) from None
        return {k: v for k, v in mapping.items() if k in columns}
    out: dict[str, Any] = {}
    for key, value in mapping.items():
        try:
            col = one(c for c in columns if snake_case(c) == snake_case(key))
        except OneEmptyError:
            raise _MapMappingToTableSnakeMapEmptyError(
                mapping=mapping, columns=columns, key=key
            ) from None
        except OneNonUniqueError as error:
            raise _MapMappingToTableSnakeMapNonUniqueError(
                mapping=mapping,
                columns=columns,
                key=key,
                first=error.first,
                second=error.second,
            ) from None
        else:
            out[col] = value
    return out


@dataclass(kw_only=True, slots=True)
class _MapMappingToTableError(Exception):
    mapping: StrMapping
    columns: Sequence[str]


@dataclass(kw_only=True, slots=True)
class _MapMappingToTableExtraColumnsError(_MapMappingToTableError):
    extra: AbstractSet[str]

    @override
    def __str__(self) -> str:
        return f"Mapping {get_repr(self.mapping)} must be a subset of table columns {get_repr(self.columns)}; got extra {self.extra}"


@dataclass(kw_only=True, slots=True)
class _MapMappingToTableSnakeMapEmptyError(_MapMappingToTableError):
    key: str

    @override
    def __str__(self) -> str:
        return f"Mapping {get_repr(self.mapping)} must be a subset of table columns {get_repr(self.columns)}; cannot find column to map to {self.key!r} modulo snake casing"


@dataclass(kw_only=True, slots=True)
class _MapMappingToTableSnakeMapNonUniqueError(_MapMappingToTableError):
    key: str
    first: str
    second: str

    @override
    def __str__(self) -> str:
        return f"Mapping {get_repr(self.mapping)} must be a subset of table columns {get_repr(self.columns)}; found columns {self.first!r}, {self.second!r} and perhaps more to map to {self.key!r} modulo snake casing"


##


def _orm_inst_to_dict(obj: DeclarativeBase, /) -> StrMapping:
    """Map an ORM instance to a dictionary."""
    attrs = {
        k for k, _ in yield_object_attributes(obj, static_type=InstrumentedAttribute)
    }
    return {
        name: _orm_inst_to_dict_one(obj, attrs, name) for name in get_column_names(obj)
    }


def _orm_inst_to_dict_one(
    obj: DeclarativeBase, attrs: AbstractSet[str], name: str, /
) -> Any:
    attr = one(
        attr for attr in attrs if _orm_inst_to_dict_predicate(type(obj), attr, name)
    )
    return getattr(obj, attr)


def _orm_inst_to_dict_predicate(
    cls: type[DeclarativeBase], attr: str, name: str, /
) -> bool:
    cls_attr = getattr(cls, attr)
    try:
        return cls_attr.name == name
    except AttributeError:
        return False


##


def _tuple_to_mapping(
    values: tuple[Any, ...], table_or_orm: TableOrORMInstOrClass, /
) -> dict[str, Any]:
    columns = get_column_names(table_or_orm)
    mapping = dict(zip(columns, tuple(values), strict=False))
    return {k: v for k, v in mapping.items() if v is not None}


__all__ = [
    "CHUNK_SIZE_FRAC",
    "CheckEngineError",
    "DialectOrEngineOrConnectionOrAsync",
    "EngineOrConnectionOrAsync",
    "ExtractURLError",
    "ExtractURLOutput",
    "GetTableError",
    "InsertItemsError",
    "TablenameMixin",
    "check_connect",
    "check_connect_async",
    "check_engine",
    "columnwise_max",
    "columnwise_min",
    "create_engine",
    "ensure_database_created",
    "ensure_database_dropped",
    "ensure_tables_created",
    "ensure_tables_dropped",
    "enum_name",
    "enum_values",
    "extract_url",
    "get_chunk_size",
    "get_column_names",
    "get_columns",
    "get_primary_key_values",
    "get_table",
    "get_table_name",
    "hash_primary_key_values",
    "insert_items",
    "is_orm",
    "is_table_or_orm",
    "migrate_data",
    "selectable_to_string",
    "yield_connection",
    "yield_primary_key_columns",
]



from __future__ import annotations

from aiohttp import InvalidUrlClientError
from pytest import mark, raises
from slack_sdk.webhook.async_client import AsyncWebhookClient

from utilities.os import get_env_var
from utilities.pytest import throttle
from utilities.slack_sdk import _get_async_client, send_to_slack, send_to_slack_async
from utilities.whenever import MINUTE


class TestGetClient:
    def test_main(self) -> None:
        client = _get_async_client("url")
        assert isinstance(client, AsyncWebhookClient)


class TestSendToSlack:
    def test_sync(self) -> None:
        with raises(ValueError, match="unknown url type"):
            send_to_slack("url", "message")

    async def test_async(self) -> None:
        with raises(InvalidUrlClientError, match="url"):
            await send_to_slack_async("url", "message")

    @mark.skipif(get_env_var("SLACK", nullable=True) is None, reason="'SLACK' not set")
    @throttle(delta=5 * MINUTE)
    async def test_real(self) -> None:
        url = get_env_var("SLACK")
        await send_to_slack_async(
            url, f"message from {TestSendToSlack.test_real.__qualname__}"
        )



from __future__ import annotations

from contextvars import ContextVar

from utilities.contextvars import (
    _GLOBAL_BREAKPOINT,
    global_breakpoint,
    set_global_breakpoint,
    yield_set_context,
)
from utilities.text import unique_str


class TestGlobalBreakpoint:
    def test_disabled(self) -> None:
        global_breakpoint()

    def test_set(self) -> None:
        try:
            set_global_breakpoint()
        finally:
            _ = _GLOBAL_BREAKPOINT.set(False)


class TestYieldSetContext:
    def test_disabled(self) -> None:
        context = ContextVar(unique_str(), default=False)
        assert not context.get()
        with yield_set_context(context):
            assert context.get()
        assert not context.get()

    def test_set(self) -> None:
        try:
            set_global_breakpoint()
        finally:
            _ = _GLOBAL_BREAKPOINT.set(False)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from itertools import chain
from os import chdir
from os.path import expandvars
from pathlib import Path
from re import IGNORECASE, search
from subprocess import PIPE, CalledProcessError, check_output
from typing import TYPE_CHECKING, Literal, assert_never, overload, override

from utilities.contextlib import enhanced_context_manager
from utilities.errors import ImpossibleCaseError
from utilities.sentinel import Sentinel

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from utilities.types import MaybeCallablePathLike, PathLike


PWD = Path.cwd()


def ensure_suffix(path: PathLike, suffix: str, /) -> Path:
    """Ensure a path has a given suffix."""
    path = Path(path)
    parts = path.name.split(".")
    suffixes = suffix.strip(".").split(".")
    max_len = max(len(parts), len(suffixes))
    try:
        i = next(i for i in range(max_len, 0, -1) if parts[-i:] == suffixes[:i])
    except StopIteration:
        add = suffixes
    else:
        add = suffixes[i:]
    name = ".".join(chain(parts, add))
    return path.with_name(name)


##


def expand_path(path: PathLike, /) -> Path:
    """Expand a path."""
    path = str(path)
    path = expandvars(path)
    return Path(path).expanduser()


##


def get_package_root(path: MaybeCallablePathLike = Path.cwd, /) -> Path:
    """Get the package root."""
    path = to_path(path)
    path_dir = path.parent if path.is_file() else path
    all_paths = list(chain([path_dir], path_dir.parents))
    try:
        return next(
            p.resolve()
            for p in all_paths
            if any(p_i.name == "pyproject.toml" for p_i in p.iterdir())
        )
    except StopIteration:
        raise GetPackageRootError(path=path) from None


@dataclass(kw_only=True, slots=True)
class GetPackageRootError(Exception):
    path: PathLike

    @override
    def __str__(self) -> str:
        return f"Path is not part of a package: {self.path}"


##


def get_repo_root(path: MaybeCallablePathLike = Path.cwd, /) -> Path:
    """Get the repo root."""
    path = to_path(path)
    path_dir = path.parent if path.is_file() else path
    try:
        output = check_output(
            ["git", "rev-parse", "--show-toplevel"],
            stderr=PIPE,
            cwd=path_dir,
            text=True,
        )
    except CalledProcessError as error:
        # newer versions of git report "Not a git repository", whilst older
        # versions report "not a git repository"
        if search("fatal: not a git repository", error.stderr, flags=IGNORECASE):
            raise GetRepoRootError(path=path) from None
        raise  # pragma: no cover
    else:
        return Path(output.strip("\n"))


@dataclass(kw_only=True, slots=True)
class GetRepoRootError(Exception):
    path: PathLike

    @override
    def __str__(self) -> str:
        return f"Path is not part of a `git` repository: {self.path}"


##


def get_root(path: MaybeCallablePathLike = Path.cwd, /) -> Path:
    """Get the root of a path."""
    path = to_path(path)
    try:
        repo = get_repo_root(path)
    except GetRepoRootError:
        repo = None
    try:
        package = get_package_root(path)
    except GetPackageRootError:
        package = None
    match repo, package:
        case None, None:
            raise GetRootError(path=path)
        case Path(), None:
            return repo
        case None, Path():
            return package
        case Path(), Path():
            if repo == package:
                return repo
            if is_sub_path(repo, package, strict=True):
                return repo
            if is_sub_path(package, repo, strict=True):
                return package
            raise ImpossibleCaseError(  # pragma: no cover
                case=[f"{repo=}", f"{package=}"]
            )
        case never:
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class GetRootError(Exception):
    path: PathLike

    @override
    def __str__(self) -> str:
        return f"Unable to determine root from {str(self.path)!r}"


##


type _GetTailDisambiguate = Literal["raise", "earlier", "later"]


def get_tail(
    path: PathLike, root: PathLike, /, *, disambiguate: _GetTailDisambiguate = "raise"
) -> Path:
    """Get the tail of a path following a root match."""
    path_parts, root_parts = [Path(p).parts for p in [path, root]]
    len_path, len_root = map(len, [path_parts, root_parts])
    if len_root > len_path:
        raise _GetTailLengthError(path=path, root=root, len_root=len_root)
    candidates = {
        i + len_root: path_parts[i : i + len_root]
        for i in range(len_path + 1 - len_root)
    }
    matches = {k: v for k, v in candidates.items() if v == root_parts}
    match len(matches), disambiguate:
        case 0, _:
            raise _GetTailEmptyError(path=path, root=root)
        case 1, _:
            return _get_tail_core(path, next(iter(matches)))
        case _, "raise":
            first, second, *_ = matches
            raise _GetTailNonUniqueError(
                path=path,
                root=root,
                first=_get_tail_core(path, first),
                second=_get_tail_core(path, second),
            )
        case _, "earlier":
            return _get_tail_core(path, next(iter(matches)))
        case _, "later":
            return _get_tail_core(path, next(iter(reversed(matches))))
        case never:
            assert_never(never)


def _get_tail_core(path: PathLike, i: int, /) -> Path:
    parts = Path(path).parts
    return Path(*parts[i:])


@dataclass(kw_only=True, slots=True)
class GetTailError(Exception):
    path: PathLike
    root: PathLike


@dataclass(kw_only=True, slots=True)
class _GetTailLengthError(GetTailError):
    len_root: int

    @override
    def __str__(self) -> str:
        return f"Unable to get the tail of {str(self.path)!r} with root of length {self.len_root}"


@dataclass(kw_only=True, slots=True)
class _GetTailEmptyError(GetTailError):
    @override
    def __str__(self) -> str:
        return (
            f"Unable to get the tail of {str(self.path)!r} with root {str(self.root)!r}"
        )


@dataclass(kw_only=True, slots=True)
class _GetTailNonUniqueError(GetTailError):
    first: Path
    second: Path

    @override
    def __str__(self) -> str:
        return f"Path {str(self.path)!r} must contain exactly one tail with root {str(self.root)!r}; got {str(self.first)!r}, {str(self.second)!r} and perhaps more"


##


def module_path(
    path: PathLike,
    /,
    *,
    root: PathLike | None = None,
    disambiguate: _GetTailDisambiguate = "raise",
) -> str:
    """Return a module path."""
    path = Path(path)
    if root is not None:
        path = get_tail(path, root, disambiguate=disambiguate)
    parts = path.with_suffix("").parts
    return ".".join(parts)


##


def is_sub_path(x: PathLike, y: PathLike, /, *, strict: bool = False) -> bool:
    """Check if a path is a sub path of another."""
    x, y = [Path(i).resolve() for i in [x, y]]
    return x.is_relative_to(y) and not (strict and y.is_relative_to(x))


##


def list_dir(path: PathLike, /) -> Sequence[Path]:
    """List the contents of a directory."""
    return sorted(Path(path).iterdir())


##


@enhanced_context_manager
def temp_cwd(path: PathLike, /) -> Iterator[None]:
    """Context manager with temporary current working directory set."""
    prev = Path.cwd()
    chdir(path)
    try:
        yield
    finally:
        chdir(prev)


##


@overload
def to_path(path: Sentinel, /) -> Sentinel: ...
@overload
def to_path(path: MaybeCallablePathLike | None = Path.cwd, /) -> Path: ...
def to_path(
    path: MaybeCallablePathLike | None | Sentinel = Path.cwd, /
) -> Path | Sentinel:
    """Get the path."""
    match path:
        case Path() | Sentinel():
            return path
        case None:
            return Path.cwd()
        case str():
            return Path(path)
        case Callable() as func:
            return to_path(func())
        case never:
            assert_never(never)


__all__ = [
    "PWD",
    "GetPackageRootError",
    "GetRepoRootError",
    "GetTailError",
    "ensure_suffix",
    "expand_path",
    "get_package_root",
    "get_repo_root",
    "get_tail",
    "is_sub_path",
    "list_dir",
    "module_path",
    "temp_cwd",
    "to_path",
]



from __future__ import annotations

from enum import Enum, auto
from functools import partial
from typing import Any

import hypothesis.strategies
from hypothesis.strategies import (
    SearchStrategy,
    booleans,
    builds,
    dictionaries,
    floats,
    integers,
    just,
    lists,
    none,
    recursive,
    sampled_from,
    tuples,
    uuids,
)

from tests.test_typing_funcs.with_future import (
    DataClassFutureCustomEquality,
    DataClassFutureDefaultInInitChild,
    DataClassFutureInt,
    DataClassFutureIntDefault,
    DataClassFutureLiteral,
    DataClassFutureLiteralNullable,
    DataClassFutureNestedInnerFirstOuter,
    DataClassFutureNestedOuterFirstOuter,
    DataClassFutureNone,
    DataClassFutureTypeLiteral,
    DataClassFutureTypeLiteralNullable,
)
from utilities.hypothesis import (
    assume_does_not_raise,
    date_deltas,
    date_periods,
    date_time_deltas,
    dates,
    int64s,
    month_days,
    paths,
    plain_date_times,
    py_datetimes,
    text_ascii,
    text_printable,
    time_deltas,
    time_periods,
    times,
    versions,
    year_months,
    zoned_date_time_periods,
    zoned_date_times,
)
from utilities.math import MAX_INT64, MIN_INT64


def objects(
    *,
    dataclass_custom_equality: bool = False,
    dataclass_default_in_init_child: bool = False,
    dataclass_int: bool = False,
    dataclass_int_default: bool = False,
    dataclass_literal: bool = False,
    dataclass_literal_nullable: bool = False,
    dataclass_nested: bool = False,
    dataclass_none: bool = False,
    dataclass_type_literal: bool = False,
    dataclass_type_literal_nullable: bool = False,
    enum: bool = False,
    exception_class: bool = False,
    exception_instance: bool = False,
    extra_base: SearchStrategy[Any] | None = None,
    floats_allow_nan: bool = False,
    sub_frozenset: bool = False,
    sub_list: bool = False,
    sub_set: bool = False,
    sub_tuple: bool = False,
    all_: bool = False,
    parsable: bool = False,
    sortable: bool = False,
) -> SearchStrategy[Any]:
    base = (
        booleans()
        | date_periods()
        | dates()
        | hypothesis.strategies.dates()
        | py_datetimes(zoned=booleans())
        | floats(allow_nan=floats_allow_nan)
        | (int64s() if parsable else integers())
        | month_days()
        | none()
        | paths()
        | plain_date_times()
        | text_printable().filter(lambda x: not x.startswith("["))
        | time_deltas()
        | time_periods()
        | times()
        | hypothesis.strategies.times()
        | uuids()
        | versions()
        | year_months()
        | zoned_date_time_periods()
        | zoned_date_times()
    )
    if dataclass_custom_equality:
        base |= builds(DataClassFutureCustomEquality)
    if dataclass_default_in_init_child:
        base |= builds(DataClassFutureDefaultInInitChild)
    if dataclass_int:
        base |= builds(DataClassFutureInt).filter(lambda obj: _is_int64(obj.int_))
    if dataclass_int_default:
        base |= builds(DataClassFutureIntDefault).filter(
            lambda obj: _is_int64(obj.int_)
        )
    if dataclass_literal:
        base |= builds(DataClassFutureLiteral, truth=sampled_from(["true", "false"]))
    if dataclass_literal_nullable:
        base |= builds(
            DataClassFutureLiteralNullable,
            truth=sampled_from(["true", "false"]) | none(),
        )
    if dataclass_nested:
        base |= builds(DataClassFutureNestedInnerFirstOuter).filter(
            lambda outer: _is_int64(outer.inner.int_)
        ) | builds(DataClassFutureNestedOuterFirstOuter).filter(
            lambda outer: _is_int64(outer.inner.int_)
        )
    if dataclass_none:
        base |= builds(DataClassFutureNone)
    if dataclass_type_literal:
        base |= builds(
            DataClassFutureTypeLiteral, truth=sampled_from(["true", "false"])
        )
    if dataclass_type_literal_nullable:
        base |= builds(
            DataClassFutureTypeLiteralNullable,
            truth=sampled_from(["true", "false"]) | none(),
        )
    if enum or all_:
        base |= sampled_from(TruthEnum)
    if exception_class or all_:
        base |= just(CustomError)
    if exception_instance or all_:
        base |= builds(CustomError, int64s())
    if extra_base is not None:
        base |= extra_base
    if not sortable:
        base |= date_deltas(parsable=parsable) | date_time_deltas(parsable=parsable)
    extend = partial(
        _extend,
        sub_frozenset=sub_frozenset,
        sub_list=sub_list,
        sub_set=sub_set,
        sub_tuple=sub_tuple,
        all_=all_,
    )
    return recursive(base, extend, max_leaves=10)


def _extend(
    strategy: SearchStrategy[Any],
    /,
    *,
    sub_frozenset: bool = False,
    sub_list: bool = False,
    sub_set: bool = False,
    sub_tuple: bool = False,
    all_: bool = False,
) -> SearchStrategy[Any]:
    lsts = lists(strategy)
    sets = lsts.map(_into_set)
    frozensets = lists(strategy).map(_into_set).map(frozenset)
    extension = (
        dictionaries(text_ascii(), strategy)
        | frozensets
        | lsts
        | sets
        | tuples(strategy)
    )
    if sub_frozenset or all_:
        extension |= frozensets.map(SubFrozenSet)
    if sub_list or all_:
        extension |= lists(strategy).map(SubList)
    if sub_set or all_:
        extension |= sets.map(SubSet)
    if sub_tuple or all_:
        extension |= tuples(strategy).map(SubTuple)
    return extension


def _is_int64(n: int, /) -> bool:
    return MIN_INT64 <= n <= MAX_INT64


def _into_set(elements: list[Any], /) -> set[Any]:
    with assume_does_not_raise(TypeError, match="unhashable type"):
        return set(elements)


class CustomError(Exception): ...


class SubFrozenSet(frozenset):
    pass


class SubList(list):
    pass


class SubSet(set):
    pass


class SubTuple(tuple):  # noqa: SLOT001
    pass


class TruthEnum(Enum):
    true = auto()
    false = auto()



from __future__ import annotations

from asyncio import sleep
from enum import Enum, StrEnum, auto
from getpass import getuser
from typing import TYPE_CHECKING, Any, Literal, assert_never, cast, override
from uuid import uuid4

from hypothesis import HealthCheck, Phase, given, settings
from hypothesis.strategies import booleans, lists, none, sets, tuples
from pytest import mark, param, raises
from sqlalchemy import (
    URL,
    Boolean,
    Column,
    Engine,
    ForeignKey,
    Integer,
    MetaData,
    Table,
    select,
)
from sqlalchemy.exc import DatabaseError, OperationalError, ProgrammingError
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    MappedAsDataclass,
    mapped_column,
    relationship,
)

from tests.conftest import SKIPIF_CI
from utilities.hypothesis import int32s, pairs, quadruples, urls
from utilities.iterables import one
from utilities.modules import is_installed
from utilities.sqlalchemy import (
    CheckEngineError,
    Dialect,
    GetTableError,
    InsertItemsError,
    TablenameMixin,
    TableOrORMInstOrClass,
    _ExtractURLDatabaseError,
    _ExtractURLHostError,
    _ExtractURLPasswordError,
    _ExtractURLPortError,
    _ExtractURLUsernameError,
    _get_dialect,
    _get_dialect_max_params,
    _insert_items_yield_merged_mappings,
    _insert_items_yield_normalized,
    _InsertItem,
    _is_pair_of_sequence_of_tuple_or_string_mapping_and_table,
    _is_pair_of_str_mapping_and_table,
    _is_pair_of_tuple_and_table,
    _is_pair_of_tuple_or_str_mapping_and_table,
    _map_mapping_to_table,
    _MapMappingToTableExtraColumnsError,
    _MapMappingToTableSnakeMapEmptyError,
    _MapMappingToTableSnakeMapNonUniqueError,
    _orm_inst_to_dict,
    _tuple_to_mapping,
    check_connect,
    check_connect_async,
    check_engine,
    columnwise_max,
    columnwise_min,
    create_engine,
    ensure_database_created,
    ensure_database_dropped,
    ensure_tables_created,
    ensure_tables_dropped,
    enum_name,
    enum_values,
    extract_url,
    get_chunk_size,
    get_column_names,
    get_columns,
    get_primary_key_values,
    get_table,
    get_table_name,
    hash_primary_key_values,
    insert_items,
    is_orm,
    is_table_or_orm,
    migrate_data,
    selectable_to_string,
    yield_primary_key_columns,
)
from utilities.text import strip_and_dedent
from utilities.typing import get_args
from utilities.whenever import MILLISECOND, get_now, to_local_plain

if TYPE_CHECKING:
    from pathlib import Path

    from utilities.types import StrMapping


def _table_names() -> str:
    """Generate at unique string."""
    key = str(uuid4()).replace("-", "")
    return f"{to_local_plain(get_now())}_{key}"


class TestCheckConnect:
    def test_sync(self, *, test_engine: Engine) -> None:
        assert check_connect(test_engine)

    async def test_async(self, *, test_async_engine: AsyncEngine) -> None:
        assert await check_connect_async(test_async_engine)

    async def test_async_timeout(self, *, test_async_engine: AsyncEngine) -> None:
        assert not await check_connect_async(test_async_engine, timeout=MILLISECOND)


class TestCheckEngine:
    async def test_main(self, *, test_async_engine: AsyncEngine) -> None:
        await check_engine(test_async_engine)

    async def test_num_tables_pass(self, *, test_async_engine: AsyncEngine) -> None:
        table = Table(
            _table_names(), MetaData(), Column("id", Integer, primary_key=True)
        )
        await ensure_tables_created(test_async_engine, table)
        match _get_dialect(test_async_engine):
            case "sqlite":
                expected = 1
            case "postgresql":
                expected = (int(1e6), 1.0)
            case _ as dialect:
                raise NotImplementedError(dialect)
        await check_engine(test_async_engine, num_tables=expected)

    async def test_num_tables_error(self, *, test_async_engine: AsyncEngine) -> None:
        with raises(CheckEngineError, match=r".* must have 100000 table\(s\); got .*"):
            await check_engine(test_async_engine, num_tables=100000)


class TestColumnwiseMinMax:
    @given(values=sets(pairs(int32s() | none()), min_size=1))
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_main(
        self,
        *,
        test_async_engine: AsyncEngine,
        values: set[tuple[int | None, int | None]],
    ) -> None:
        table = Table(
            _table_names(),
            MetaData(),
            Column("id_", Integer, primary_key=True, autoincrement=True),
            Column("x", Integer),
            Column("y", Integer),
        )
        await insert_items(
            test_async_engine, ([{"x": x, "y": y} for x, y in values], table)
        )
        sel = select(
            table.c["x"],
            table.c["y"],
            columnwise_min(table.c["x"], table.c["y"]).label("min_xy"),
            columnwise_max(table.c["x"], table.c["y"]).label("max_xy"),
        )
        async with test_async_engine.begin() as conn:
            res = (await conn.execute(sel)).all()
        assert len(res) == len(values)
        for x, y, min_xy, max_xy in res:
            if (x is None) and (y is None):
                assert min_xy is None
                assert max_xy is None
            elif (x is not None) and (y is None):
                assert min_xy == x
                assert max_xy == x
            elif (x is None) and (y is not None):
                assert min_xy == y
                assert max_xy == y
            else:
                assert min_xy == min(x, y)
                assert max_xy == max(x, y)

    async def test_label(self, *, test_async_engine: AsyncEngine) -> None:
        table = Table(
            _table_names(),
            MetaData(),
            Column("id_", Integer, primary_key=True, autoincrement=True),
            Column("x", Integer),
        )
        await ensure_tables_created(test_async_engine, table)
        sel = select(columnwise_min(table.c.x, table.c.x))
        async with test_async_engine.begin() as conn:
            _ = (await conn.execute(sel)).all()


class TestCreateEngine:
    @mark.parametrize(
        ("drivername", "async_", "cls"),
        [param("sqlite", False, Engine), param("sqlite+aiosqlite", True, AsyncEngine)],
    )
    @mark.parametrize(
        "query", [param({"arg1": "value1", "arg2": ["value2"]}), param(None)]
    )
    def test_main(
        self,
        *,
        drivername: str,
        tmp_path: Path,
        async_: bool,
        query: StrMapping | None,
        cls: type[Any],
    ) -> None:
        engine = create_engine(
            drivername, database=tmp_path.name, query=query, async_=async_
        )
        assert isinstance(engine, cls)


@SKIPIF_CI
class TestEnsureDatabaseCreatedAndDropped:
    async def test_main(self) -> None:
        url = URL.create(
            "postgresql+asyncpg",
            username=getuser(),
            password="password",  # noqa: S106
            host="localhost",
            port=5432,
            database="postgres",
        )
        database = f"testing_{_table_names()}"
        for _ in range(2):
            await ensure_database_created(url, database)
            await sleep(0.1)
        for _ in range(2):
            await ensure_database_dropped(url, database)
            await sleep(0.1)

    async def _run_test(
        self, engine: AsyncEngine, table_or_orm: TableOrORMInstOrClass, /
    ) -> None:
        for _ in range(2):
            await ensure_tables_created(engine, table_or_orm)
        sel = select(get_table(table_or_orm))
        async with engine.begin() as conn:
            _ = (await conn.execute(sel)).all()


class TestEnsureTablesCreated:
    async def test_table(self, *, test_async_engine: AsyncEngine) -> None:
        table = Table(
            _table_names(), MetaData(), Column("id_", Integer, primary_key=True)
        )
        await self._run_test(test_async_engine, table)

    async def test_mapped_class(self, *, test_async_engine: AsyncEngine) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        await self._run_test(test_async_engine, Example)

    async def _run_test(
        self, engine: AsyncEngine, table_or_orm: TableOrORMInstOrClass, /
    ) -> None:
        for _ in range(2):
            await ensure_tables_created(engine, table_or_orm)
        sel = select(get_table(table_or_orm))
        async with engine.begin() as conn:
            _ = (await conn.execute(sel)).all()


class TestEnsureTablesDropped:
    async def test_table(self, *, test_async_engine: AsyncEngine) -> None:
        table = Table(
            _table_names(), MetaData(), Column("id_", Integer, primary_key=True)
        )
        await self._run_test(test_async_engine, table)

    async def test_mapped_class(self, *, test_async_engine: AsyncEngine) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        await self._run_test(test_async_engine, Example)

    async def _run_test(
        self, engine: AsyncEngine, table_or_orm: TableOrORMInstOrClass, /
    ) -> None:
        for _ in range(2):
            await ensure_tables_dropped(engine, table_or_orm)
        sel = select(get_table(table_or_orm))
        with raises(DatabaseError):
            async with engine.begin() as conn:
                _ = await conn.execute(sel)


class TestEnumName:
    def test_main(self) -> None:
        class Example(Enum): ...

        result = enum_name(Example)
        assert result == "example_enum"


class TestEnumValues:
    def test_main(self) -> None:
        class Example(StrEnum):
            true = auto()
            false = auto()

        result = enum_values(Example)
        expected = ["true", "false"]
        assert result == expected


class TestExtractURL:
    @given(url=urls(all_=True))
    def test_main(self, *, url: URL) -> None:
        extracted = extract_url(url)
        assert extracted.username == url.username
        assert extracted.password == url.password
        assert extracted.host == url.host
        assert extracted.port == url.port
        assert extracted.database == url.database

    @given(url=urls(username=False))
    def test_username(self, *, url: URL) -> None:
        with raises(
            _ExtractURLUsernameError,
            match="Expected URL to contain a user name; got .*",
        ):
            _ = extract_url(url)

    @given(url=urls(username=True, password=False))
    def test_password(self, *, url: URL) -> None:
        with raises(
            _ExtractURLPasswordError, match="Expected URL to contain a password; got .*"
        ):
            _ = extract_url(url)

    @given(url=urls(username=True, password=True, host=False))
    def test_host(self, *, url: URL) -> None:
        with raises(
            _ExtractURLHostError, match="Expected URL to contain a host; got .*"
        ):
            _ = extract_url(url)

    @given(url=urls(username=True, password=True, host=True, port=False))
    def test_port(self, *, url: URL) -> None:
        with raises(
            _ExtractURLPortError, match="Expected URL to contain a port; got .*"
        ):
            _ = extract_url(url)

    @given(url=urls(username=True, password=True, host=True, port=True, database=False))
    def test_database(self, *, url: URL) -> None:
        with raises(
            _ExtractURLDatabaseError, match="Expected URL to contain a database; got .*"
        ):
            _ = extract_url(url)


class TestGetChunkSize:
    @mark.parametrize(
        ("num_cols", "chunk_size_frac", "expected"),
        [
            param(2, 1.0, 50),
            param(2, 0.5, 25),
            param(10, 1.0, 10),
            param(10, 0.5, 5),
            param(100, 1.0, 1),
            param(100, 0.5, 1),
        ],
    )
    def test_table(
        self, *, num_cols: int, chunk_size_frac: float, expected: int
    ) -> None:
        table = Table(
            _table_names(),
            MetaData(),
            *[Column(f"id{i}", Integer) for i in range(num_cols)],
        )
        result = get_chunk_size("sqlite", table, chunk_size_frac=chunk_size_frac)
        assert result == expected


class TestGetColumnNames:
    def test_table(self) -> None:
        table = Table(
            _table_names(), MetaData(), Column("id_", Integer, primary_key=True)
        )
        self._run_test(table)

    def test_mapped_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        self._run_test(Example)

    def _run_test(self, table_or_orm: TableOrORMInstOrClass, /) -> None:
        assert get_column_names(table_or_orm) == ["id_"]


class TestGetColumns:
    def test_table(self) -> None:
        table = Table(
            _table_names(), MetaData(), Column("id", Integer, primary_key=True)
        )
        self._run_test(table)

    def test_mapped_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        self._run_test(Example)

    def _run_test(self, table_or_orm: TableOrORMInstOrClass, /) -> None:
        columns = get_columns(table_or_orm)
        assert isinstance(columns, list)
        assert len(columns) == 1
        assert isinstance(columns[0], Column)


class TestGetDialect:
    @mark.skipif(condition=not is_installed("pyodbc"), reason="'pyodbc' not installed")
    def test_mssql(self) -> None:
        engine = create_engine("mssql")
        assert _get_dialect(engine) == "mssql"

    @mark.skipif(
        condition=not is_installed("mysqldb"), reason="'mysqldb' not installed"
    )
    def test_mysql(self) -> None:
        engine = create_engine("mysql")
        assert _get_dialect(engine) == "mysql"

    @mark.skipif(
        condition=not is_installed("oracledb"), reason="'oracledb' not installed"
    )
    def test_oracle(self) -> None:
        engine = create_engine("oracle+oracledb")
        assert _get_dialect(engine) == "oracle"

    @mark.skipif(
        condition=not is_installed("asyncpg"), reason="'asyncpg' not installed"
    )
    def test_postgres(self) -> None:
        engine = create_engine("postgresql+asyncpg")
        assert _get_dialect(engine) == "postgresql"

    @mark.skipif(
        condition=not is_installed("aiosqlite"), reason="'asyncpg' not installed"
    )
    def test_sqlite(self) -> None:
        engine = create_engine("sqlite+aiosqlite")
        assert _get_dialect(engine) == "sqlite"


class TestGetDialectMaxParams:
    @mark.parametrize("dialect", get_args(Dialect))
    def test_max_params(self, *, dialect: Dialect) -> None:
        max_params = _get_dialect_max_params(dialect)
        assert isinstance(max_params, int)


class TestGetPrimaryKeyValues:
    @given(id1=int32s(), id2=int32s(), value=booleans())
    def test_main(self, *, id1: int, id2: int, value: bool) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id1: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)
            id2: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)
            value: Mapped[bool] = mapped_column(Boolean, kw_only=True, nullable=False)

        obj = Example(id1=id1, id2=id2, value=value)
        result = get_primary_key_values(obj)
        expected = (id1, id2)
        assert result == expected


class TestGetTable:
    def test_table(self) -> None:
        table = Table(
            _table_names(), MetaData(), Column("id_", Integer, primary_key=True)
        )
        result = get_table(table)
        assert result is table

    def test_mapped_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        result = get_table(Example)
        expected = Example.__table__
        assert result is expected

    def test_instance_of_mapped_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        obj = Example(id=1)
        result = get_table(obj)
        expected = Example.__table__
        assert result is expected

    def test_error(self) -> None:
        with raises(
            GetTableError, match="Object .* must be a Table or mapped class; got .*"
        ):
            _ = get_table(cast("Any", type(None)))


class TestGetTableName:
    def test_table(self) -> None:
        name = _table_names()
        table = Table(name, MetaData(), Column("id_", Integer, primary_key=True))
        result = get_table_name(table)
        assert result == name

    def test_mapped_class(self) -> None:
        name = _table_names()

        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = name

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        result = get_table_name(Example)
        assert result == name


class TestHashPrimaryKeyValues:
    @given(id1=int32s(), id2=int32s(), value=booleans())
    def test_main(self, *, id1: int, id2: int, value: bool) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id1: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)
            id2: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)
            value: Mapped[bool] = mapped_column(Boolean, kw_only=True, nullable=False)

            @override
            def __hash__(self) -> int:
                return hash_primary_key_values(self)

        obj = Example(id1=id1, id2=id2, value=value)
        result = hash(obj)
        expected = hash((id1, id2))
        assert result == expected


class TestInsertItems:
    # end-to-end

    @given(id_=int32s())
    @mark.parametrize("case", [param("tuple"), param("dict")])
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_insert__pair_of_obj_and_table(
        self,
        *,
        case: Literal["tuple", "dict"],
        id_: int,
        test_async_engine: AsyncEngine,
    ) -> None:
        table = self._make_table()
        match case:
            case "tuple":
                item = (id_,), table
            case "dict":
                item = {"id_": id_}, table
            case never:
                assert_never(never)
        await self._run_insert_and_select(test_async_engine, table, {id_}, item)

    @given(ids=sets(int32s(), min_size=1))
    @mark.parametrize(
        "case",
        [
            param("pair-list-of-dicts"),
            param("pair-list-of-tuples"),
            param("list-of-pair-of-dicts"),
            param("list-of-pair-of-tuples"),
        ],
    )
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_insert__pair_of_objs_and_table_or_list_of_pairs_of_objs_and_table(
        self,
        *,
        case: Literal[
            "pair-list-of-tuples",
            "pair-list-of-dicts",
            "list-of-pair-of-tuples",
            "list-of-pair-of-dicts",
        ],
        ids: set[int],
        test_async_engine: AsyncEngine,
    ) -> None:
        table = self._make_table()
        match case:
            case "pair-list-of-tuples":
                item = [((id_,)) for id_ in ids], table
            case "pair-list-of-dicts":
                item = [({"id_": id_}) for id_ in ids], table
            case "list-of-pair-of-tuples":
                item = [((id_,), table) for id_ in ids]
            case "list-of-pair-of-dicts":
                item = [({"id_": id_}, table) for id_ in ids]
            case never:
                assert_never(never)
        await self._run_insert_and_select(test_async_engine, table, ids, item)

    @given(ids=sets(int32s(), min_size=10, max_size=100))
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_insert__many_items(
        self, *, ids: set[int], test_async_engine: AsyncEngine
    ) -> None:
        table = self._make_table()
        await self._run_insert_and_select(
            test_async_engine, table, ids, [({"id_": id_}, table) for id_ in ids]
        )

    @given(id_=int32s())
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_insert__mapped_class(
        self, *, id_: int, test_async_engine: AsyncEngine
    ) -> None:
        cls = self._make_mapped_class()
        await self._run_insert_and_select(test_async_engine, cls, {id_}, cls(id_=id_))

    @given(ids=sets(int32s(), min_size=1))
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_insert__mapped_classes(
        self, *, ids: set[int], test_async_engine: AsyncEngine
    ) -> None:
        cls = self._make_mapped_class()
        await self._run_insert_and_select(
            test_async_engine, cls, ids, [cls(id_=id_) for id_ in ids]
        )

    @given(id_=int32s())
    @mark.parametrize("key", [param("Id_"), param("id_")])
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_insert__snake(
        self, *, id_: int, key: str, test_async_engine: AsyncEngine
    ) -> None:
        table = self._make_table(title=True)
        item = {key: id_}, table
        await self._run_insert_and_select(
            test_async_engine, table, {id_}, item, snake=True
        )

    @given(id_=int32s(), init=booleans() | none(), post=booleans() | none())
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_insert__upsert_single_column_one_row(
        self,
        *,
        id_: int,
        init: bool | None,
        test_async_engine: AsyncEngine,
        post: bool | None,
    ) -> None:
        table = self._make_table(value=True)
        init_item = {"id_": id_, "value": init}, table
        await insert_items(test_async_engine, init_item, is_upsert=True)
        sel = select(table.c.value)
        async with test_async_engine.begin() as conn:
            assert (await conn.execute(sel)).scalar_one() == init
        post_item = {"id_": id_, "value": post}, table
        await insert_items(test_async_engine, post_item, is_upsert=True)
        async with test_async_engine.begin() as conn:
            assert (await conn.execute(sel)).scalar_one() == (
                init if post is None else post
            )

    @given(
        ids=pairs(int32s(), unique=True, sorted=True),
        init=pairs(booleans() | none()),
        post=pairs(booleans() | none()),
    )
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_insert__upsert_single_column_two_rows(
        self,
        *,
        ids: tuple[int, int],
        init: tuple[bool | None, bool | None],
        test_async_engine: AsyncEngine,
        post: tuple[bool | None, bool | None],
    ) -> None:
        table = self._make_table(value=True)
        init_items = [
            ({"id_": id_, "value": init_i}, table)
            for id_, init_i in zip(ids, init, strict=True)
        ]
        await insert_items(test_async_engine, *init_items, is_upsert=True)
        sel = select(table.c.value).order_by(table.c.id_)
        async with test_async_engine.begin() as conn:
            assert (await conn.execute(sel)).scalars().all() == list(init)
        post_items = [
            ({"id_": id_, "value": post_i}, table)
            for id_, post_i in zip(ids, post, strict=True)
        ]
        await insert_items(test_async_engine, *post_items, is_upsert=True)
        async with test_async_engine.begin() as conn:
            assert (await conn.execute(sel)).scalars().all() == [
                (init_i if post_i is None else post_i)
                for init_i, post_i in zip(init, post, strict=True)
            ]

    @given(
        id_=int32s(), init=pairs(booleans() | none()), post=pairs(booleans() | none())
    )
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_insert__upsert_two_columns_one_row(
        self,
        *,
        id_: int,
        test_async_engine: AsyncEngine,
        init: tuple[bool | None, bool | None],
        post: tuple[bool | None, bool | None],
    ) -> None:
        table = self._make_table(two_values=True)
        init_item = {"id_": id_, "value1": init[0], "value2": init[1]}, table
        await insert_items(test_async_engine, init_item, is_upsert=True)
        sel = select(table.c.value1, table.c.value2)
        async with test_async_engine.begin() as conn:
            assert (await conn.execute(sel)).one() == (init[0], init[1])
        post_item = {"id_": id_, "value1": post[0], "value2": post[1]}, table
        await insert_items(test_async_engine, post_item, is_upsert=True)
        async with test_async_engine.begin() as conn:
            assert (await conn.execute(sel)).one() == (
                init[0] if post[0] is None else post[0],
                init[1] if post[1] is None else post[1],
            )

    @given(
        ids=pairs(int32s(), unique=True, sorted=True),
        init=quadruples(booleans() | none()),
        post=quadruples(booleans() | none()),
    )
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_insert__upsert_two_columns_two_rows(
        self,
        *,
        ids: tuple[int, int],
        init: tuple[bool | None, bool | None, bool | None, bool | None],
        test_async_engine: AsyncEngine,
        post: tuple[bool | None, bool | None, bool | None, bool | None],
    ) -> None:
        table = self._make_table(two_values=True)
        init_items = [
            ({"id_": ids[0], "value1": init[0], "value2": init[1]}, table),
            ({"id_": ids[1], "value1": init[2], "value2": init[3]}, table),
        ]
        await insert_items(test_async_engine, *init_items, is_upsert=True)
        sel = select(table.c.value1, table.c.value2).order_by(table.c.id_)
        async with test_async_engine.begin() as conn:
            assert (await conn.execute(sel)).all() == [
                (init[0], init[1]),
                (init[2], init[3]),
            ]
        post_items = [
            ({"id_": ids[0], "value1": post[0], "value2": post[1]}, table),
            ({"id_": ids[1], "value1": post[2], "value2": post[3]}, table),
        ]
        await insert_items(test_async_engine, *post_items, is_upsert=True)
        async with test_async_engine.begin() as conn:
            assert (await conn.execute(sel)).all() == [
                (
                    init[0] if post[0] is None else post[0],
                    init[1] if post[1] is None else post[1],
                ),
                (
                    init[2] if post[2] is None else post[2],
                    init[3] if post[3] is None else post[3],
                ),
            ]

    @given(id_=int32s())
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_insert__assume_table_exists(
        self, *, id_: int, test_async_engine: AsyncEngine
    ) -> None:
        table = self._make_table()
        with raises(
            (OperationalError, ProgrammingError), match="(no such table|does not exist)"
        ):
            await insert_items(
                test_async_engine, ({"id_": id_}, table), assume_tables_exist=True
            )

    # yield merged mappings

    def test_yield_merged_mappings__id_and_value(self) -> None:
        table = self._make_table(value=True)
        items = [
            {"id_": 1, "value": True},
            {"id_": 1, "value": False},
            {"id_": 2, "value": False},
            {"id_": 2, "value": True},
        ]
        result = list(_insert_items_yield_merged_mappings(table, items))
        expected = [{"id_": 1, "value": False}, {"id_": 2, "value": True}]
        assert result == expected

    def test_yield_merged_mappings__value_only(self) -> None:
        table = self._make_table(value=True)
        items = [{"value": 1}, {"value": 2}]
        result = list(_insert_items_yield_merged_mappings(table, items))
        assert result == items

    def test_yield_merged_mappings__autoincrement(self) -> None:
        table = Table(
            _table_names(),
            MetaData(),
            Column("id_", Integer, primary_key=True, autoincrement=True),
            Column("value", Integer),
        )
        items = [{"value": 1}, {"value": 2}]
        result = list(_insert_items_yield_merged_mappings(table, items))
        assert result == items

    # yield normalized

    @given(id_=int32s())
    @mark.parametrize("case", [param("tuple"), param("dict")])
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    def test_yield_normalized__pair_of_tuple_or_str_mapping_and_table(
        self, *, case: Literal["tuple", "dict"], id_: int
    ) -> None:
        table = self._make_table()
        match case:
            case "tuple":
                item = (id_,), table
            case "dict":
                item = {"id_": id_}, table
            case never:
                assert_never(never)
        result = one(_insert_items_yield_normalized(item))
        expected = (table, {"id_": id_})
        assert result == expected

    @given(ids=sets(int32s(), min_size=1))
    @mark.parametrize("case", [param("tuple"), param("dict")])
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    def test_yield_normalized__pair_of_list_of_tuples_or_str_mappings_and_table(
        self, *, case: Literal["tuple", "dict"], ids: set[int]
    ) -> None:
        table = self._make_table()
        match case:
            case "tuple":
                item = [((id_,)) for id_ in ids], table
            case "dict":
                item = [({"id_": id_}) for id_ in ids], table
            case never:
                assert_never(never)
        result = list(_insert_items_yield_normalized(item))
        expected = [(table, {"id_": id_}) for id_ in ids]
        assert result == expected

    @given(ids=sets(int32s(), min_size=1))
    @mark.parametrize("case", [param("tuple"), param("dict")])
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    def test_yield_normalize__list_of_pairs_of_objs_and_table(
        self, *, case: Literal["tuple", "dict"], ids: set[int]
    ) -> None:
        table = self._make_table()
        match case:
            case "tuple":
                item = [(((id_,), table)) for id_ in ids]
            case "dict":
                item = [({"id_": id_}, table) for id_ in ids]
            case never:
                assert_never(never)
        result = list(_insert_items_yield_normalized(item))
        expected = [(table, {"id_": id_}) for id_ in ids]
        assert result == expected

    @given(id_=int32s())
    def test_yield_normalized__mapped_class(self, *, id_: int) -> None:
        cls = self._make_mapped_class()
        result = one(_insert_items_yield_normalized(cls(id_=id_)))
        expected = (get_table(cls), {"id_": id_})
        assert result == expected

    @given(ids=sets(int32s(), min_size=1))
    def test_yield_normalized__mapped_classes(self, *, ids: set[int]) -> None:
        cls = self._make_mapped_class()
        result = list(_insert_items_yield_normalized([cls(id_=id_) for id_ in ids]))
        expected = [(get_table(cls), {"id_": id_}) for id_ in ids]
        assert result == expected

    @given(id_=int32s())
    @mark.parametrize("case", [param("tuple"), param("dict")])
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    def test_yield_normalize__snake(
        self, *, case: Literal["tuple", "dict"], id_: int
    ) -> None:
        table = self._make_table(title=True)
        match case:
            case "tuple":
                item = (id_,), table
            case "dict":
                item = {"id_": id_}, table
            case never:
                assert_never(never)
        result = one(_insert_items_yield_normalized(item, snake=True))
        expected = (table, {"Id_": id_})
        assert result == expected

    @mark.parametrize(
        "item",
        [
            param((None,), id="tuple, not pair"),
            param(
                (None, Table(_table_names(), MetaData())),
                id="pair, first element invalid",
            ),
            param(((1, 2, 3), None), id="pair, second element invalid"),
            param([None], id="iterable, invalid"),
            param(None, id="outright invalid"),
        ],
    )
    def test_errors(self, *, item: Any) -> None:
        with raises(InsertItemsError, match="Item must be valid; got .*"):
            _ = list(_insert_items_yield_normalized(item))

    # private

    def _make_table(
        self, *, title: bool = False, value: bool = False, two_values: bool = False
    ) -> Table:
        return Table(
            _table_names(),
            MetaData(),
            Column("Id_" if title else "id_", Integer, primary_key=True),
            *([Column("value", Boolean, nullable=True)] if value else []),
            *(
                [Column(f"value{i}", Boolean, nullable=True) for i in [1, 2]]
                if two_values
                else []
            ),
        )

    def _make_mapped_class(self) -> type[DeclarativeBase]:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        return Example

    async def _run_insert_and_select(
        self,
        engine: AsyncEngine,
        table_or_orm: TableOrORMInstOrClass,
        ids: set[int],
        /,
        *items: _InsertItem,
        snake: bool = False,
    ) -> None:
        await insert_items(engine, *items, snake=snake)
        sel = select(get_table(table_or_orm).c["Id_" if snake else "id_"])
        async with engine.begin() as conn:
            results = (await conn.execute(sel)).scalars().all()
        assert set(results) == ids


class TestIsPairOfSequenceOfTupleOrStringMappingAndTable:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(None, False),
            param(([(1, 2, 3)], Table(_table_names(), MetaData())), True),
            param(
                ([{"a": 1, "b": 2, "c": 3}], Table(_table_names(), MetaData())), True
            ),
            param(
                (
                    [(1, 2, 3), {"a": 1, "b": 2, "c": 3}],
                    Table(_table_names(), MetaData()),
                ),
                True,
            ),
        ],
    )
    def test_main(self, *, obj: Any, expected: bool) -> None:
        result = _is_pair_of_sequence_of_tuple_or_string_mapping_and_table(obj)
        assert result is expected


class TestIsPairOfStrMappingAndTable:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(None, False),
            param(((1, 2, 3), Table(_table_names(), MetaData())), False),
            param(({"a": 1, "b": 2, "c": 3}, Table(_table_names(), MetaData())), True),
        ],
    )
    def test_main(self, *, obj: Any, expected: bool) -> None:
        result = _is_pair_of_str_mapping_and_table(obj)
        assert result is expected


class TestIsPairOfTupleAndTable:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(None, False),
            param(((1, 2, 3), Table(_table_names(), MetaData())), True),
            param(({"a": 1, "b": 2, "c": 3}, Table(_table_names(), MetaData())), False),
        ],
    )
    def test_main(self, *, obj: Any, expected: bool) -> None:
        result = _is_pair_of_tuple_and_table(obj)
        assert result is expected


class TestIsPairOfTupleStrMappingAndTable:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(None, False),
            param(((1, 2, 3), Table(_table_names(), MetaData())), True),
            param(({"a": 1, "b": 2, "c": 3}, Table(_table_names(), MetaData())), True),
        ],
    )
    def test_main(self, *, obj: Any, expected: bool) -> None:
        result = _is_pair_of_tuple_or_str_mapping_and_table(obj)
        assert result is expected


class TestIsORM:
    def test_orm_inst(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        obj = Example(id_=1)
        assert is_table_or_orm(obj)

    def test_orm_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        assert is_table_or_orm(Example)

    def test_table(self) -> None:
        table = Table(
            _table_names(), MetaData(), Column("id_", Integer, primary_key=True)
        )
        assert not is_orm(table)

    def test_none(self) -> None:
        assert not is_orm(None)


class TestIsTableOrORM:
    def test_table(self) -> None:
        table = Table(
            _table_names(), MetaData(), Column("id_", Integer, primary_key=True)
        )
        assert is_table_or_orm(table)

    def test_orm_inst(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        obj = Example(id_=1)
        assert is_table_or_orm(obj)

    def test_orm_class(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        assert is_table_or_orm(Example)

    def test_other(self) -> None:
        assert not is_table_or_orm(None)


class TestMapMappingToTable:
    @given(id_=int32s(), value=booleans())
    def test_main(self, *, id_: int, value: bool) -> None:
        mapping = {"id_": id_, "value": value}
        table = Table(
            _table_names(),
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean),
        )
        result = _map_mapping_to_table(mapping, table)
        assert result == mapping

    @given(id_=int32s(), value=booleans())
    @mark.parametrize("key1", [param("Id_"), param("id_")])
    @mark.parametrize("key2", [param("Value"), param("value")])
    def test_snake(self, *, key1: str, id_: int, key2: str, value: bool) -> None:
        mapping = {key1: id_, key2: value}
        table = Table(
            _table_names(),
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean),
        )
        result = _map_mapping_to_table(mapping, table, snake=True)
        expected = {"id_": id_, "value": value}
        assert result == expected

    @given(id_=int32s(), value=booleans(), extra=booleans())
    def test_error_extra_columns(self, *, id_: int, value: bool, extra: bool) -> None:
        mapping = {"id_": id_, "value": value, "extra": extra}
        table = Table(
            _table_names(),
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean),
        )
        with raises(
            _MapMappingToTableExtraColumnsError,
            match=r"Mapping .* must be a subset of table columns .*; got extra .*",
        ):
            _ = _map_mapping_to_table(mapping, table)

    @given(id_=int32s(), value=booleans())
    def test_error_snake_empty_error(self, *, id_: int, value: bool) -> None:
        mapping = {"id_": id_, "invalid": value}
        table = Table(
            _table_names(),
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean),
        )
        with raises(
            _MapMappingToTableSnakeMapEmptyError,
            match=r"Mapping .* must be a subset of table columns .*; cannot find column to map to 'invalid' modulo snake casing",
        ):
            _ = _map_mapping_to_table(mapping, table, snake=True)

    @given(id_=int32s(), value=booleans())
    def test_error_snake_non_unique_error(self, *, id_: int, value: bool) -> None:
        mapping = {"id_": id_, "value": value}
        table = Table(
            _table_names(),
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean),
            Column("Value", Boolean),
        )
        with raises(
            _MapMappingToTableSnakeMapNonUniqueError,
            match=r"Mapping .* must be a subset of table columns .*; found columns 'value', 'Value' and perhaps more to map to 'value' modulo snake casing",
        ):
            _ = _map_mapping_to_table(mapping, table, snake=True)


class TestMigrateData:
    @given(
        values=lists(
            tuples(int32s(), booleans() | none()), min_size=1, unique_by=lambda x: x[0]
        )
    )
    @settings(
        max_examples=1,
        phases={Phase.generate},
        suppress_health_check={HealthCheck.function_scoped_fixture},
    )
    async def test_main(
        self, *, values: list[tuple[int, bool]], test_async_engine: AsyncEngine
    ) -> None:
        table1 = self._make_table()
        await insert_items(
            test_async_engine, [({"id_": id_, "value": v}, table1) for id_, v in values]
        )
        async with test_async_engine.begin() as conn:
            result1 = (await conn.execute(select(table1))).all()
        assert len(result1) == len(values)

        table2 = self._make_table()
        await migrate_data(
            table1, test_async_engine, test_async_engine, table_or_orm_to=table2
        )
        async with test_async_engine.begin() as conn:
            result2 = (await conn.execute(select(table2))).all()
        assert len(result2) == len(values)

    def _make_table(self) -> Table:
        return Table(
            _table_names(),
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean, nullable=True),
        )


class TestORMInstToDict:
    @given(id_=int32s())
    def test_main(self, *, id_: int) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        example = Example(id_=id_)
        result = _orm_inst_to_dict(example)
        expected = {"id_": id_}
        assert result == expected

    @given(id_=int32s())
    def test_explicitly_named_column(self, *, id_: int) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Example(Base):
            __tablename__ = _table_names()

            ID: Mapped[int] = mapped_column(
                Integer, kw_only=True, primary_key=True, name="id"
            )

        example = Example(ID=id_)
        result = _orm_inst_to_dict(example)
        expected = {"id": id_}
        assert result == expected

    @given(parent_id=int32s(), child_id=int32s())
    def test_relationship(self, *, parent_id: int, child_id: int) -> None:
        class Base(DeclarativeBase, MappedAsDataclass): ...

        class Parent(Base):
            __tablename__ = _table_names()

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

            children: Mapped[list[Child]] = relationship(
                "Child", init=False, back_populates="parent"
            )

        class Child(Base):
            __tablename__ = _table_names()

            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)
            parent_id: Mapped[int] = mapped_column(
                ForeignKey(f"{Parent.__tablename__}.id_"), kw_only=True, nullable=False
            )

            parent: Mapped[Parent] = relationship(
                "Parent", init=False, back_populates="children"
            )

        parent = Parent(id_=parent_id)
        result = _orm_inst_to_dict(parent)
        expected = {"id_": parent_id}
        assert result == expected

        child = Child(id_=child_id, parent_id=parent_id)
        result = _orm_inst_to_dict(child)
        expected = {"id_": child_id, "parent_id": parent_id}
        assert result == expected


class TestSelectableToString:
    async def test_main(self, *, test_async_engine: AsyncEngine) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean, nullable=True),
        )
        sel = select(table).where(table.c.value >= 1)
        result = selectable_to_string(sel, test_async_engine)
        expected = strip_and_dedent(
            """
                SELECT example.id_, example.value *
                FROM example *
                WHERE example.value >= 1
            """.replace("*", "")
        )
        assert result == expected


class TestTablenameMixin:
    def test_main(self) -> None:
        class Base(DeclarativeBase, MappedAsDataclass, TablenameMixin): ...

        class Example(Base):
            id_: Mapped[int] = mapped_column(Integer, kw_only=True, primary_key=True)

        assert get_table_name(Example) == "example"


class TestTupleToMapping:
    @mark.parametrize(
        ("values", "expected"),
        [
            param((), {}),
            param((1,), {"id_": 1}),
            param((1, True), {"id_": 1, "value": True}),
            param((None, True), {"value": True}),
        ],
        ids=str,
    )
    def test_main(self, *, values: tuple[Any, ...], expected: StrMapping) -> None:
        table = Table(
            _table_names(),
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", Boolean, nullable=True),
        )
        result = _tuple_to_mapping(values, table)
        assert result == expected


class TestYieldPrimaryKeyColumns:
    def test_main(self) -> None:
        table = Table(
            _table_names(),
            MetaData(),
            Column("id1", Integer, primary_key=True),
            Column("id2", Integer, primary_key=True),
            Column("id3", Integer),
        )
        result = list(yield_primary_key_columns(table))
        expected = [
            Column("id1", Integer, primary_key=True),
            Column("id2", Integer, primary_key=True),
        ]
        for c, e in zip(result, expected, strict=True):
            assert c.name == e.name
            assert c.primary_key == e.primary_key

    def test_autoincrement(self) -> None:
        table = Table(
            _table_names(),
            MetaData(),
            Column("id_", Integer, primary_key=True, autoincrement=True),
            Column("x", Integer),
            Column("y", Integer),
        )
        result = list(yield_primary_key_columns(table, autoincrement=False))
        assert result == []



from __future__ import annotations

from dataclasses import dataclass, field
from math import isclose, nan
from typing import TYPE_CHECKING, Self

from psutil import swap_memory, virtual_memory

from utilities.contextlib import suppress_super_object_attribute_error
from utilities.whenever import get_now

if TYPE_CHECKING:
    from whenever import ZonedDateTime


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class MemoryUsage:
    """A memory usage."""

    datetime: ZonedDateTime = field(default_factory=get_now)
    virtual_used: int = field(repr=False)
    virtual_used_mb: int = field(init=False)
    virtual_total: int = field(repr=False)
    virtual_total_mb: int = field(init=False)
    virtual_pct: float = field(init=False)
    swap_used: int = field(repr=False)
    swap_used_mb: int = field(init=False)
    swap_total: int = field(repr=False)
    swap_total_mb: int = field(init=False)
    swap_pct: float = field(init=False)

    def __post_init__(self) -> None:
        with suppress_super_object_attribute_error():
            super().__post_init__()  # pyright: ignore[reportAttributeAccessIssue]
        self.virtual_used_mb = self._to_mb(self.virtual_used)
        self.virtual_total_mb = self._to_mb(self.virtual_total)
        self.virtual_pct = (
            nan
            if isclose(self.virtual_total, 0.0)
            else self.virtual_used / self.virtual_total
        )
        self.swap_used_mb = self._to_mb(self.swap_used)
        self.swap_total_mb = self._to_mb(self.swap_total)
        self.swap_pct = (
            nan if isclose(self.swap_total, 0.0) else self.swap_used / self.swap_total
        )

    @classmethod
    def new(cls) -> Self:
        virtual = virtual_memory()
        virtual_total = virtual.total
        swap = swap_memory()
        return cls(
            virtual_used=virtual_total - virtual.available,
            virtual_total=virtual_total,
            swap_used=swap.used,
            swap_total=swap.total,
        )

    def _to_mb(self, bytes_: int) -> int:
        return round(bytes_ / (1024**2))


__all__ = ["MemoryUsage"]
